@@ -110,10 +110,15 @@ module:hook_tag(xmlns_sasl, "failure", function (session, stanza)
 	module:log("info", "SASL EXTERNAL with %s failed: %s", session.to_host, condition);
 
 	session.external_auth = "failed"
-	session:close();
-	return true;
+	session.external_auth_failure_reason = condition;
 end, 500)
 
+module:hook_tag(xmlns_sasl, "failure", function (session, stanza) -- luacheck: ignore 212/stanza
+	session.log("debug", "No fallback from SASL EXTERNAL failure, giving up");
+	session:close(nil, session.external_auth_failure_reason);
+	return true;
+end, 90)
+
 module:hook_tag("http://etherx.jabber.org/streams", "features", function (session, stanza)
 	if session.type ~= "s2sout_unauthed" or not session.secure then return; end
 
@@ -263,6 +268,7 @@ module:hook("stream-features", function(event)
 			elseif not origin.secure and insecure_mechanisms:contains(mechanism) then
 				log("debug", "Not offering mechanism %s on insecure connection", mechanism);
 			else
+				log("debug", "Offering mechanism %s", mechanism);
 				mechanisms:tag("mechanism"):text(mechanism):up();
 			end
 		end
