@@ -358,6 +358,7 @@ end
 module:hook("muc-disco#info", function(event)
 	event.reply:tag("feature", {var = "http://jabber.org/protocol/muc"}):up();
 	event.reply:tag("feature", {var = "http://jabber.org/protocol/muc#stable_id"}):up();
+	event.reply:tag("feature", {var = "http://jabber.org/protocol/muc#self-ping-optimization"}):up();
 end);
 module:hook("muc-disco#info", function(event)
 	table.insert(event.form, { name = "muc#roominfo_occupants", label = "Number of occupants" });
@@ -523,6 +524,9 @@ function room_mt:handle_normal_presence(origin, stanza)
 			log("debug", "no occupant found for %s; creating new occupant object for %s", dest_jid, real_jid);
 			is_first_dest_session = true;
 			dest_occupant = self:new_occupant(bare_jid, dest_jid);
+			if orig_occupant then
+				dest_occupant.role = orig_occupant.role;
+			end
 		else
 			is_first_dest_session = false;
 		end
@@ -727,7 +731,7 @@ function room_mt:handle_iq_to_occupant(origin, stanza)
 	else -- Type is "get" or "set"
 		local current_nick = self:get_occupant_jid(from);
 		if not current_nick then
-			origin.send(st.error_reply(stanza, "cancel", "not-acceptable"));
+			origin.send(st.error_reply(stanza, "cancel", "not-acceptable", "You are not currently connected to this chat"));
 			return true;
 		end
 		if not occupant then -- recipient not in room
@@ -760,7 +764,7 @@ function room_mt:handle_message_to_occupant(origin, stanza)
 	local type = stanza.attr.type;
 	if not current_nick then -- not in room
 		if type ~= "error" then
-			origin.send(st.error_reply(stanza, "cancel", "not-acceptable"));
+			origin.send(st.error_reply(stanza, "cancel", "not-acceptable", "You are not currently connected to this chat"));
 		end
 		return true;
 	end
@@ -778,6 +782,7 @@ function room_mt:handle_message_to_occupant(origin, stanza)
 		return true;
 	end
 	log("debug", "%s sent private message stanza to %s (%s)", from, to, o_data.jid);
+	stanza = muc_util.filter_muc_x(st.clone(stanza));
 	stanza:tag("x", { xmlns = "http://jabber.org/protocol/muc#user" }):up();
 	stanza.attr.from = current_nick;
 	self:route_to_occupant(o_data, stanza)
@@ -1057,7 +1062,7 @@ end
 module:hook("muc-occupant-groupchat", function(event)
 	local role_rank = valid_roles[event.occupant and event.occupant.role or "none"];
 	if role_rank <= valid_roles.none then
-		event.origin.send(st.error_reply(event.stanza, "cancel", "not-acceptable"));
+		event.origin.send(st.error_reply(event.stanza, "cancel", "not-acceptable", "You are not currently connected to this chat"));
 		return true;
 	elseif role_rank <= valid_roles.visitor then
 		event.origin.send(st.error_reply(event.stanza, "auth", "forbidden"));
