@@ -25,7 +25,8 @@ local log = require "util.logger".init("sasl");
 local generate_uuid = require "util.uuid".generate;
 local nodeprep = require "util.encodings".stringprep.nodeprep;
 
-module "sasl.digest-md5"
+local _ENV = nil;
+-- luacheck: std none
 
 --=========================
 --SASL DIGEST-MD5 according to RFC 2831
@@ -241,8 +242,10 @@ local function digest(self, message)
 	end
 end
 
-function init(registerMechanism)
+local function init(registerMechanism)
 	registerMechanism("DIGEST-MD5", {"plain"}, digest);
 end
 
-return _M;
+return {
+	init = init;
+}
